import time
import threading
import aprslib
import logging
from aprslib.parsing import parse
from aprslib.exceptions import ParseError, UnknownFormat
from queue import Queue, Empty

YIELD_DELAY = 0.001
POLL_DELAY = 0.1

logger = logging.getLogger("aprstastic")


class APRSClient(object):
    """
    A thin, thread-safe, client around aprslib.
    Maybe we will implement our own one day!
    """

    def __init__(self, login: str, passcode: str, filters: None):
        super().__init__()
        self._login = login
        self._passcode = passcode
        self._filters = filters
        self._aprs = None
        self._rx_queue = Queue()
        self._tx_queue = Queue()

        self._rx_thread = threading.Thread(target=self._rx_thread_body)
        self._tx_thread = threading.Thread(target=self._tx_thread_body)
        self._rx_thread.start()
        self._tx_thread.start()

    def recv(self, raw=False) -> str | None:
        """
        Returns one packet from the receive queue, or None if the queue is empty."
        """
        try:
            packet = self._rx_queue.get(block=False)
            if raw:
                return packet
            else:
                try:
                    return parse(packet)
                except ParseError:
                    logger.error("ParseError: " + packet.strip())
                except UnknownFormat:
                    logger.error("UnknownFormat: " + packet.strip())
        except Empty:
            return None

    def send(self, packet: str) -> None:
        """
        Enqueue a packet on the send queue, to be sent ASAP.
        """
        self._tx_queue.put(packet)

    def set_filter(self, filters: str | None) -> None:
        """
        Update the filters controling which packets are received from APRS IS
        """
        self._tx_queue.put(_UpdateFilters(filters))

    def _tx_thread_body(self) -> None:
        while True:
            try:
                # Client has not been created yet
                if self._aprs is None:
                    time.sleep(POLL_DELAY)
                    continue

                # Client is not yet connected
                if not self._aprs._connected:
                    time.sleep(POLL_DELAY)
                    continue

                # Read a packet
                packet = self._tx_queue.get(block=False)
                if isinstance(packet, _UpdateFilters):
                    self._filters = packet.filters
                    if self._filters is None:
                        self._aprs.set_filter("")  # Default filter?
                    else:
                        self._aprs.set_filter(self._filters)
                else:
                    self._aprs.sendall(packet)
            except Empty:
                time.sleep(YIELD_DELAY)

    def _rx_thread_body(self) -> None:
        self._aprs = aprslib.IS(self._login, passwd=self._passcode, port=14580)
        if self._filters is not None:
            self._aprs.set_filter(self._filters)
        self._aprs.connect()
        self._aprs.consumer(
            lambda x: self._rx_queue.put(x, block=True), raw=True, blocking=True
        )


class _UpdateFilters(object):
    """
    Class used to update the filters.
    """

    def __init__(self, filters: str):
        super().__init__()
        self.filters = filters



# SPDX-FileCopyrightText: 2024-present Adam Fourney <adam.fourney@gmail.com>
#
# SPDX-License-Identifier: MIT
import logging
import yaml
import os
import sys
from logging.handlers import TimedRotatingFileHandler
from ._config import init_config, ConfigError
from ._gateway import Gateway

# Set up logging
################
logger = logging.getLogger("aprstastic")
logging.root.setLevel(logging.DEBUG)


class LocalDebugFilter(logging.Filter):
    """
    Only show debug messages if they are from arpstastic
    """

    def filter(self, record):
        if record.name == "aprstastic":
            return True
        else:
            return record.levelno > logging.DEBUG


stream_handler = logging.StreamHandler()
stream_handler.setFormatter(logging.Formatter(logging.BASIC_FORMAT))
stream_handler.addFilter(LocalDebugFilter())
logging.root.addHandler(stream_handler)

# Load the configuration
config = None
try:
    config = init_config()
except ConfigError as e:
    sys.stderr.write(str(e).rstrip() + "\n")
    sys.exit(1)

# Configure file logs
logs_dir = config.get("gateway", {}).get("logs_dir")
if logs_dir is not None:
    log_file = os.path.join(logs_dir, "aprstastic.log")
    logger.debug(f"Writing logs to: {log_file}")
    file_handler = TimedRotatingFileHandler(
        log_file, when="d", interval=1, backupCount=7
    )
    file_handler.setFormatter(logging.Formatter(logging.BASIC_FORMAT))
    file_handler.addFilter(LocalDebugFilter())
    logging.root.addHandler(file_handler)

# Start the gateway
gateway = Gateway(config)
gateway.run()



# SPDX-FileCopyrightText: 2024-present Adam Fourney <adam.fourney@gmail.com>
#
# SPDX-License-Identifier: MIT
import sqlite3
import logging
import json
import shutil
import time
import requests
import traceback
import os

from .__about__ import __version__

logger = logging.getLogger("aprstastic")

DATABASE_FILE = "registrations.db"

# The precompiled registrations and overrides are not updated directly, and so are loaded from flat files
OVERRIDES_FILE = "registration_overrides.json"
PRECOMPILED_FILE = "precompiled_registrations.json"


class CallSignRegistry(object):
    """
    There are three potential sources of registration data:
        - local (via direct Meshtastic messages to this device)
        - beaconed (observing registration beacons to APRS)
        - precompiled (a precompiled database, downloaded from the web, to seed things)

    There is also a system-level registration which is meant to override everything if needed.

    The records are then merged, by date (with system-level overrides having the final word).
    This class manages this process
    """

    def __init__(self, data_dir):
        super().__init__()
        self._data_dir = data_dir

        self._db_conn = self._open_db(os.path.join(data_dir, DATABASE_FILE))
        self._precompiled = self._load_precompiled(
            os.path.join(data_dir, PRECOMPILED_FILE)
        )
        self._overrides = self._load_overrides(os.path.join(data_dir, OVERRIDES_FILE))
        self._merged = dict()

        self._rebuild()

    def _open_db(self, db_path):
        """
        Return a sqlite database connection to the registration database, initilizing the database if needed.
        """
        create_db = not os.path.isfile(db_path)
        conn = sqlite3.connect(db_path)
        cursor = conn.cursor()

        if create_db:
            cursor.execute(
                """
CREATE TABLE IF NOT EXISTS VersionInfo (
    db_version INTEGER,
    package_version TEXT
)
"""
            )

            cursor.execute(
                """
CREATE TABLE IF NOT EXISTS LocalRegistrations (
    device_id TEXT UNIQUE,
    call_sign TEXT UNIQUE,
    timestamp INTEGER
)
"""
            )

            cursor.execute(
                """
CREATE TABLE IF NOT EXISTS BeaconedRegistrations (
    device_id TEXT UNIQUE,
    call_sign TEXT UNIQUE,
    timestamp INTEGER
)
"""
            )

            cursor.execute(
                "INSERT INTO VersionInfo (db_version, package_version) VALUES (?, ?);",
                (1, __version__),
            )

            # Commit the changes
            conn.commit()
            logger.debug(f"initialized database: {db_path}")

        cursor.close()
        return conn

    def add_registration(self, device_id, call_sign, is_local):
        cursor = self._db_conn.cursor()

        # Delete prior rows
        del_query = "DELETE FROM %s WHERE device_id = ? OR call_sign = ?;" % (
            "LocalRegistrations" if is_local else "BeaconedRegistrations",
        )
        cursor.execute(del_query, (device_id, call_sign))

        # Insert the new record
        insert_query = (
            "INSERT INTO %s (device_id, call_sign, timestamp) VALUES (?, ?, ?);"
            % ("LocalRegistrations" if is_local else "BeaconedRegistrations",)
        )
        cursor.execute(insert_query, (device_id, call_sign, int(time.time())))

        self._db_conn.commit()
        cursor.close()

        self._rebuild()

    def _rebuild(self):
        """
        Updates (by rebuilding), the in-memory copy of the merged database, applying all rules.
        """
        cursor = self._db_conn.cursor()
        self._merged = dict()
        self._merged.update(self._precompiled)

        beaconed_registrations = dict()
        cursor.execute(
            "SELECT device_id, call_sign, timestamp FROM BeaconedRegistrations;"
        )
        rows = cursor.fetchall()
        for row in rows:
            if row[0] in self._merged:  # Check the timestamps
                if self._merged[row[0]][1] > row[2]:
                    continue
            self._merged[row[0]] = (row[1], row[2])

        local_registrations = dict()
        cursor.execute(
            "SELECT device_id, call_sign, timestamp FROM LocalRegistrations;"
        )
        rows = cursor.fetchall()
        for row in rows:
            if row[0] in self._merged:  # Check the timestamps
                if self._merged[row[0]][1] > row[2]:
                    continue
            self._merged[row[0]] = (row[1], row[2])

        self._merged.update(self._overrides)

        # Strip un-needed data
        for k in self._merged:
            self._merged[k] = self._merged[k][0]

        cursor.close()

    def _load_overrides(self, file_path):
        """
        Return a copy of the registration overrides, which is loaded into memory.
        """
        return dict()

    def _load_precompiled(self, file_path):
        """
        Return a copy of the precompiled registrations, which are loaded into memory.
        """

        # Doesn't exist, so copy the version that shipped with this app
        if not os.path.isfile(file_path):
            packaged_database = os.path.join(
                os.path.dirname(os.path.abspath(__file__)), "res", PRECOMPILED_FILE
            )
            logger.debug(
                f"Copying packaged version of the precompiled database: {packaged_database}"
            )
            shutil.copyfile(packaged_database, file_path)

        # Load the existing copy
        precompiled_data = None
        with open(file_path, "rt") as fh:
            precompiled_data = json.loads(fh.read())
        cached_timestamp = precompiled_data.get("download_timestamp", 0)
        now = time.time()

        # It's older than a day, so download a new one
        if now - cached_timestamp > 3600 * 24:
            logger.debug("Downloading precompiled database.")
            try:
                response = requests.get(precompiled_data.get("url"))
                response.raise_for_status()
                precompiled_data = json.loads(response.text)
                precompiled_data["download_timestamp"] = now
                precompiled_data["reported_timestamp"] = min(
                    now, precompiled_data["reported_timestamp"]
                )
            except:
                logger.error(traceback.format_exc())
            # Save it
            with open(file_path, "wt") as fh:
                fh.write(json.dumps(precompiled_data, indent=4))

        # Return tuples in the expected format
        result = dict()
        for t in precompiled_data["tuples"]:
            result[t[0]] = (t[1], min(now, t[2]))  # No future timestamps
        return result

    # Emulate a dictionary
    def __getitem__(self, key):
        return self._merged[key]

    def __len__(self):
        return len(self._merged)

    def keys(self):
        return self._merged.keys()

    def values(self):
        return self._merged.values()

    def items(self):
        return self._merged.items()

    def __contains__(self, item):
        return item in self._merged

    def __iter__(self):
        return iter(self._merged)



import os
import yaml
import logging
from ._config_const import DEFAULT_CONFIG_CONTENT

logger = logging.getLogger("aprstastic")

CONFIG_FILE_NAME = "aprstastic.yaml"
LOGS_SUBDIR = "logs"
DATA_SUBDIR = "data"
DEFAULT_DATA_DIR = os.path.join(os.path.expanduser("~"), ".config", "aprstastic")
DEFAULT_CALL_SIGN = "N0CALL"


class ConfigError(Exception):
    pass


def init_config():
    """
    Return a valid gateway configuration, or raise an error.
    This method performs the following steps:

        1. Find the config.yaml file. If unsuccessful, create a sample, and exit.
        2. Validate the configuration.
        3. Determine the data directory -- creating it if it does not exist.
        4. Determine the logging directory -- creating it if it does not exits.
    """

    # Find the config file
    config_path = None
    if os.path.isfile(CONFIG_FILE_NAME):
        config_path = CONFIG_FILE_NAME
    else:
        path = os.path.join(DEFAULT_DATA_DIR, CONFIG_FILE_NAME)
        if os.path.isfile(path):
            config_path = path
        else:
            os.makedirs(DEFAULT_DATA_DIR, exist_ok=True)
            with open(os.path.join(DEFAULT_DATA_DIR, CONFIG_FILE_NAME), "wt") as fh:
                fh.write(DEFAULT_CONFIG_CONTENT.rstrip() + "\n")
            raise ConfigError(
                f"ERROR: No '{CONFIG_FILE_NAME}' file was found.\nA sample configuration was written to '{path}'\nBE SURE TO EDIT THIS FILE BEFORE PROCEEDING.\nAt the very least, edit the 'call_sign' and 'aprsis_passcode' values!"
            )

    # Load the config file
    config = None
    logger.debug(f"config file: {config_path}")
    with open(config_path, "r") as file:
        config = yaml.safe_load(file)

    # Validate
    # TODO

    # Make sure the call sign was at least changed
    call_sign = config.get("call_sign")
    if call_sign == DEFAULT_CALL_SIGN:
        raise ConfigError(
            f"ERROR: '{config_path}' appears to be the sample config.\nBE SURE TO EDIT THIS FILE BEFORE PROCEEDING.\nAt the very least, edit the 'call_sign' and 'aprsis_passcode' values!"
        )

    # Try to get the data directory
    data_dir = config.get("data_dir")
    if data_dir is None:
        data_dir = os.path.join(
            os.path.dirname(os.path.abspath(config_path)), DATA_SUBDIR
        )
    elif data_dir.strip() == "":
        data_dir = os.path.join(os.getcwd(), DATA_SUBDIR)
    else:
        data_dir = os.path.expanduser(data_dir)

    logger.debug(f"data directory: {data_dir}")
    os.makedirs(data_dir, exist_ok=True)
    config["data_dir"] = data_dir

    # Try to get the logging directory
    logs_dir = config.get("logs_dir")
    if logs_dir is None:
        logs_dir = os.path.join(
            os.path.dirname(os.path.abspath(config_path)), LOGS_SUBDIR
        )
    elif logs_dir.strip() == "":
        logs_dir = os.path.join(os.getcwd(), LOGS_SUBDIR)
    else:
        logs_dir = os.path.expanduser(logs_dir)

    logger.debug(f"logs directory: {logs_dir}")
    os.makedirs(logs_dir, exist_ok=True)
    config["logs_dir"] = logs_dir

    # TODO: Fixme
    # Wrap the config in a structure that is backwards compatible
    return {"gateway": config, "licensed_operators": {}}



import aprslib
from aprslib.parsing import parse
import pubsub
import time
import sys
import json
import time
import logging
import random
import threading
import re
import os
import serial.tools.list_ports
import meshtastic.stream_interface
import meshtastic.serial_interface
from datetime import datetime
from meshtastic.util import message_to_json

from queue import Queue, Empty
from .__about__ import __version__
from ._aprs_client import APRSClient
from ._registry import CallSignRegistry

logger = logging.getLogger("aprstastic")

from aprslib.parsing import parse

MQTT_TOPIC = "meshtastic.receive"
REGISTRATION_BEACON = "MESHID-01"


class Gateway(object):
    def __init__(self, config):
        self._config = config

        self._gateway_id = None
        self._gateway_call_sign = None

        self._interface = None
        self._mesh_rx_queue = Queue()

        self._aprs_client = None

        self._reply_to = {}
        self._filtered_call_signs = []
        self._beacon_registrations = False

        self._id_to_call_signs = CallSignRegistry(
            config.get("gateway", {}).get("data_dir")
        )

    def run(self):
        # Connect to the Meshtastic device
        device = (
            self._config.get("gateway", {})
            .get("meshtastic_interface", {})
            .get("device")
        )

        self._interface = self._get_interface(device)
        if self._interface is None:
            raise ValueError("No meshtastic device detected or specified.")

        def on_recv(packet):
            self._mesh_rx_queue.put(packet)

        pubsub.pub.subscribe(on_recv, MQTT_TOPIC)
        node_info = self._interface.getMyNodeInfo()
        self._gateway_id = node_info.get("user", {}).get("id")
        logger.debug(f"Gateway device id: {self._gateway_id}")

        # Create an initial list of known call signs based on the device node database

        # Myself
        self._gateway_call_sign = (
            self._config.get("gateway", {}).get("call_sign", "").upper().strip()
        )
        self._filtered_call_signs.append(self._gateway_call_sign)

        # The registraion beacon
        self._beacon_registrations = self._config.get("gateway", {}).get(
            "beacon_registrations", True
        )
        if self._beacon_registrations:
            self._filtered_call_signs.append(REGISTRATION_BEACON)

        # Recently seen nodes
        for node in self._interface.nodesByNum.values():
            presumptive_id = f"!{node['num']:08x}"

            if presumptive_id not in self._id_to_call_signs:
                continue

            # Heard more than a day ago
            last_heard = node.get("lastHeard")
            if last_heard is None or last_heard + 3600 * 24 < time.time():
                continue

            self._filtered_call_signs.append(self._id_to_call_signs[presumptive_id])

        # Connect to APRS IS
        aprsis_passcode = self._config.get("gateway", {}).get("aprsis_passcode")
        self._aprs_client = APRSClient(
            self._gateway_call_sign,
            aprsis_passcode,
            "g/" + "/".join(self._filtered_call_signs),
        )

        logger.debug("Pausing for 2 seconds...")
        time.sleep(2.0)
        logger.debug("Starting main loop.")

        while True:
            # Read a meshastic packet
            mesh_packet = None
            try:
                mesh_packet = self._mesh_rx_queue.get(block=False)
            except Empty:
                pass

            if mesh_packet is not None:
                self._process_meshtastic_packet(mesh_packet)

            # Read an APRS packet
            aprs_packet = self._aprs_client.recv()
            if aprs_packet is not None:
                self._process_aprs_packet(aprs_packet)

            # Yield
            time.sleep(0.001)

    def _get_interface(
        self, device=None
    ) -> meshtastic.stream_interface.StreamInterface:
        if device is None:
            ports = list(serial.tools.list_ports.comports())
            if len(ports) == 1:
                device = ports[0].device
        if device is not None:
            dev = meshtastic.serial_interface.SerialInterface(device)
            logger.info(f"Connected to: {device}")
            return dev
        else:
            return None

    def _process_meshtastic_packet(self, packet):
        fromId = packet.get("fromId", None)
        toId = packet.get("toId", None)
        portnum = packet.get("decoded", {}).get("portnum")

        # Don't bother logging my telemetry
        if portnum != "TELEMETRY_APP" or fromId != self._gateway_id:
            logger.info(f"{fromId} -> {toId}: {portnum}")

        # Record that we have spotted the ID
        should_announce = self._spotted(fromId)

        if portnum == "POSITION_APP":
            if fromId not in self._id_to_call_signs:
                return

            position = packet.get("decoded", {}).get("position")
            self._send_aprs_position(
                self._id_to_call_signs[fromId],
                position.get("latitude"),
                position.get("longitude"),
                position.get("time"),
                "aprstastic: " + fromId,
            )

        if portnum == "TEXT_MESSAGE_APP":
            message_bytes = packet["decoded"]["payload"]
            message_string = message_bytes.decode("utf-8")

            if toId == "^all" and message_string.lower().strip().startswith("aprs?"):
                self._send_mesh_message(
                    fromId,
                    "APRS-tastic Gateway available here. Welcome. Reply '?' for more info.",
                )
                return

            if toId != self._gateway_id:
                # Not for me
                return

            if message_string.strip() == "?":
                # It seems to send in the opposite order? LIFO?
                self._send_mesh_message(
                    fromId,
                    "Send and receive APRS messages by registering your call sign. HAM license required.\n\nReply with:\n!register [CALLSIGN]-[SSID]\nE.g.,\n!register N0CALL-1\n\nSee https://github.com/afourney/aprstastic for more.",
                )
                return

            if message_string.strip() == "!id" or message_string.strip() == "!version":
                # Let clients query the gateway call sign and version number
                self._send_mesh_message(
                    fromId,
                    f"Gateway call sign: {self._gateway_call_sign}, Version: {__version__}",
                )
                return

            if message_string.lower().strip().startswith("!register"):
                # Allow operatores to join
                m = re.search(
                    r"^!register:?\s+([a-z0-9]{4,7}\-[0-9]{1,2})$",
                    message_string.lower().strip(),
                )
                if m:
                    call_sign = m.group(1).upper()
                    if fromId in self._id_to_call_signs:
                        # Update
                        self._id_to_call_signs.add_registration(fromId, call_sign, True)
                        self._send_mesh_message(fromId, "Registration updated.")
                    else:
                        # New
                        self._id_to_call_signs.add_registration(fromId, call_sign, True)
                        self._spotted(fromId)
                        self._send_mesh_message(
                            fromId,
                            "Registered. Send APRS messages by replying here, and prefixing your message with the dest callsign. E.g., 'WLNK-1: hello' ",
                        )
                    self._spotted(fromId)  # Run this again to update subscriptions

                    # Beacon the registration to APRS-IS to facilitate building a shared roaming mapping
                    # which will be queried in future updates to aprstastic.
                    if self._beacon_registrations:
                        logger.info(
                            f"Beaconing registration {call_sign} <-> {fromId}, to {REGISTRATION_BEACON}"
                        )
                        self._send_aprs_message(call_sign, REGISTRATION_BEACON, fromId)
                else:
                    self._send_mesh_message(
                        fromId,
                        "Invalid call sign + ssid.\nSYNTAX: !register [CALLSIGN]-[SSID]\nE.g.,\n!register N0CALL-1",
                    )
                return

            if fromId not in self._id_to_call_signs:
                self._send_mesh_message(
                    fromId,
                    "Unknown device. HAM license required!\nRegister by replying with:\n!register [CALLSIGN]-[SSID]\nE.g.,\n!register N0CALL-1",
                )
                return

            m = re.search(r"^([A-Za-z0-9]+(\-[A-Za-z0-9]+)?):(.*)$", message_string)
            if m:
                tocall = m.group(1)
                self._reply_to[fromId] = tocall
                message = m.group(3).strip()
                self._send_aprs_message(self._id_to_call_signs[fromId], tocall, message)
                return
            elif fromId in self._reply_to:
                self._send_aprs_message(
                    self._id_to_call_signs[fromId],
                    self._reply_to[fromId],
                    message_string,
                )
                return
            else:
                self._send_mesh_message(
                    fromId,
                    "Please prefix your message with the dest callsign. E.g., 'WLNK-1: hello'",
                )
                return

        # At this point the message was not handled yet. Announce yourself
        if should_announce:
            self._send_mesh_message(
                fromId,
                "APRS-tastic Gateway available here. Welcome. Reply '?' for more info.",
            )

    def _process_aprs_packet(self, packet):
        if packet.get("format") == "message":
            fromcall = packet.get("from", "N0CALL").strip().upper()
            tocall = packet.get("addresse", "").strip().upper()

            # Is this an ack?
            if packet.get("response") == "ack":
                logger.debug(
                    f"Received ACK to {tocall}'s message #" + packet.get("msgNo", "")
                )
                return

            # Is this a registration beacon?
            if tocall == REGISTRATION_BEACON:
                mesh_id = packet.get("message_text", "").lower().strip()
                if re.search(r"^![a-f0-9]{8}$", mesh_id):
                    logger.info(
                        f"Observed registration beacon: {mesh_id}: {fromcall}",
                    )
                    self._id_to_call_signs.add_registration(mesh_id, fromcall, False)
                else:
                    logger.error(
                        f"Invalid registration beacon: {packet.get('raw')}",
                    )
                return

            # Ack all remaining messages (which aren't themselves acks, and aren't beacons)
            self._send_aprs_ack(tocall, fromcall, packet.get("msgNo", ""))

            # Message was sent to the gateway itself. Print it. Do nothing else.
            if tocall == self._gateway_call_sign:
                logger.info(
                    f"Received APRS message addressed to the gateway ({self._gateway_call_sign}): {packet['raw']}"
                )
                return

            # Figure out where the packet is going
            toId = None
            for k in self._id_to_call_signs:
                if tocall == self._id_to_call_signs[k].strip().upper():
                    toId = k
                    break
            if toId is None:
                loggin.error(f"Unkown recipient: {tocall}")
                return

            # Forward the message
            message = packet.get("message_text")
            if message is not None:
                self._reply_to[toId] = fromcall
                self._send_mesh_message(toId, fromcall + ": " + message)

    def _send_aprs_message(self, fromcall, tocall, message):
        while len(tocall) < 9:
            tocall += " "
        packet = (
            fromcall
            + ">APRS,WIDE1-1,qAR,"
            + self._gateway_call_sign
            + "::"
            + tocall
            + ":"
            + message.strip()
            + "{"
            + str(random.randint(0, 999))
        )
        logger.debug("Sending to APRS: " + packet)
        self._aprs_client.send(packet)

    def _send_aprs_ack(self, fromcall, tocall, messageId):
        while len(tocall) < 9:
            tocall += " "
        packet = (
            fromcall
            + ">APRS,WIDE1-1,qAR,"
            + self._gateway_call_sign
            + "::"
            + tocall
            + ":ack"
            + messageId
        )
        logger.debug("Sending to APRS: " + packet)
        self._aprs_client.send(packet)

    def _send_aprs_position(self, fromcall, lat, lon, t, message):
        aprs_lat_ns = "N" if lat >= 0 else "S"
        lat = abs(lat)
        aprs_lat_deg = int(lat)
        aprs_lat_min = float((lat - aprs_lat_deg) * 60)
        aprs_lat = f"%02d%02.2f%s" % (aprs_lat_deg, aprs_lat_min, aprs_lat_ns)

        aprs_lon_ns = "E" if lon >= 0 else "W"
        lon = abs(lon)
        aprs_lon_deg = abs(int(lon))
        aprs_lon_min = float((lon - aprs_lon_deg) * 60)
        aprs_lon = f"%03d%05.2f%s" % (aprs_lon_deg, aprs_lon_min, aprs_lon_ns)

        aprs_ts = datetime.utcfromtimestamp(t).strftime("%d%H%M")
        if len(aprs_ts) == 5:
            aprs_ts = "0" + aprs_ts + "z"
        else:
            aprs_ts = aprs_ts + "z"

        aprs_msg = "@" + aprs_ts + aprs_lat + "/" + aprs_lon + ">" + message
        packet = (
            fromcall + ">APRS,WIDE1-1,qAR," + self._gateway_call_sign + ":" + aprs_msg
        )
        logger.debug(f"Sending to APRS: {packet}")
        self._aprs_client.send(packet)

    def _send_mesh_message(self, destid, message):
        logger.info(f"Sending to '{destid}': {message}")
        self._interface.sendText(
            text=message, destinationId=destid, wantAck=True, wantResponse=False
        )

    def _spotted(self, node_id):
        """
        Checks if the node is registered, and newly spotted -- in which
        case we need to update the APRS filters, and also greet the user.
        In this case, it returns true. Otherwise, false.
        """

        # We spotted them, but they aren't registered
        if node_id not in self._id_to_call_signs:
            return False

        call_sign = self._id_to_call_signs[node_id]

        if call_sign in self._filtered_call_signs:
            return False

        # Ok it's new, update the filters
        self._filtered_call_signs.append(call_sign)
        self._aprs_client.set_filter("g/" + "/".join(self._filtered_call_signs))
        return True



DEFAULT_CONFIG_CONTENT = """# 
# APRSTASTIC CONFIGURATION FILE (version: 1)
# Be sure to at least modify 'call_sign' and 'aprsis_passcode'.
#


# Radio call sign of the gateway itself (analogy, iGate's call sign)
call_sign: N0CALL 


# APRS-IS passcode. Search Google for how to get this
aprsis_passcode: 12345 


# Only serial devices are supported right now. 
# If 'device' is null (or commented out), an attempt will be made to 
# detected it automatically.
meshtastic_interface: 
   type: serial
#  device: /dev/ttyACM0


# Beacon new registrations to APRS-IS to facilitate discovery
beacon_registrations: true 


# Where should logs be stored?
# If null, (or commented out), store logs in the `logs` dir, sibling to this file. 
#log_dir: null


# Where should data be stored?
# If null, (or commented out), store data in the `data` dir, sibling to this file. 
#data_dir: null
"""
